import torch
import torch.nn as nn
import torch.nn.functional as F


class ResidualBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride=1):
        super(ResidualBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=in_channels,
                                out_channels=out_channels,
                                kernel_size=(3, 3),
                                stride=stride,
                                padding=1,
                                bias=False)
        
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.conv2 = nn.Conv2d(in_channels=out_channels,
                                out_channels=out_channels,
                                kernel_size=(3, 3),
                                stride=1,
                                padding=1,
                                bias=False)
        
        self.bn2 = nn.BatchNorm2d(out_channels)
        self.shortcut = nn.Sequential()
        if stride != 1 or in_channels != out_channels:
            self.shortcut = nn.Sequential(nn.Conv2d(in_channels=in_channels,
                                                    out_channels=out_channels,
                                                    kernel_size=(1, 1),
                                                    stride=stride,
                                                    bias=False),
                                          nn.BatchNorm2d(out_channels))
    def forward(self, x):
        out = F.relu((self.bn1(self.conv1(x))))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, num_classes=10, pretrained=True):
        super(ResNet, self).__init__()        
        self.conv1 = nn.Conv2d(in_channels=3,
                                out_channels=64,
                                kernel_size=(3, 3),
                                stride=1,
                                padding=1,
                                bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.block1 = self._create_block(64, 64, stride=1)
        self.block2 = self._create_block(64, 128, stride=8)
        self.linear = nn.Linear(128, num_classes)
        self.avgPool = nn.AdaptiveAvgPool2d(1)
        
        if(pretrained):
            self.load_checkpoint()
    
    def _create_block(self, in_channels, out_channels, stride):
        return nn.Sequential(ResidualBlock(in_channels, out_channels, stride),
                             ResidualBlock(out_channels, out_channels, 1))
    
    def load_checkpoint(self):
        checkpoint = '../checkpoints/resnet_cifar10.730.chkpt'
        try:
            self.load_state_dict(torch.load(checkpoint))
        except Exception as error:
            print(f'There was a problem attempting to load the checkpoint {checkpoint}')
            print(error)

    def forward(self, x):
        out = F.relu((self.bn1(self.conv1(x))))
        out = self.block1(out)
        out = self.block2(out)
        out = self.avgPool(out)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out