import torch
import torch.nn as nn
import torch.nn.functional as F
from resnetSmall import ResNet, ResidualBlock

class Cifar10ResNet(torch.nn.Module):
    """
    This is a wrapper class for our custom ResNet model. Its purpose is to
    preprocess the input tensors and ensure that all tensors are created on
    the correct device with the correct dtype. It also defines the variable
    userLabel as an attribute that can be controlled by a custom knob in Nuke.
    """
    def __init__(self, userLabel = 0):
        """
        :param userLabel: This int controls which object is detected by the
                          Cifar10ResNet model and can be linked to an enumeration
                          knob in Nuke.
        """
        super(Cifar10ResNet, self).__init__()
        self.model = ResNet(pretrained=True)
        self.model.eval()
        self.userLabel = userLabel

    def normalize(self, input, mean, std):
        """
        This method normalizes the values in input based on mean and std.
        :param input: a torch.Tensor of the size [batch x 3 x H x W]
        :param mean: A tuple of float values that represent the mean of the
                            r,g,b chans e.g. [0.5, 0.5, 0.5]
        :param std: A tuple of float values that represent the std of the
                          r,g,b chans e.g. [0.5, 0.5, 0.5]
        :return: a torch.Tensor that has been normalised
        """
        # type: (Tensor, Tuple[float, float, float], Tuple[float, float, float]) -> Tensor
        input[:, 0, :, :] = (input[:, 0, :, :] - mean[0]) / std[0]
        input[:, 1, :, :] = (input[:, 1, :, :] - mean[1]) / std[1]
        input[:, 2, :, :] = (input[:, 2, :, :] - mean[2]) / std[2]
        return input

    def forward(self, input):
        """
        The forward function for this model will normalize the input and then pass it
        to the Resnet forward function. It will compare the returned label to userLabel label and
        return either an all ones or all zeros tensor.
        :param input: A torch.Tensor of size 1 x 3 x H x W representing the input image
        :return: A torch.Tensor of size 1 x 1 x H x W of zeros or ones
        """
        # Determine which device all tensors should be created on
        if(input.is_cuda):
          device = torch.device('cuda')
        else:
          device = torch.device('cpu')

        # Normalise the input tensor
        mean = (0.5, 0.5, 0.5)
        std = (0.5, 0.5, 0.5)
        input = self.normalize(input, mean, std)

        modelOutput = self.model.forward(input)
        modelLabel = int(torch.argmax(modelOutput[0]))

        # Check if the detected object is the same as userLabel
        if modelLabel == self.userLabel:
          # Ensure output is created on the correct device with the correct dtype
          output = torch.ones((1, 1, input.shape[2], input.shape[3]), dtype = input.dtype, device=device)
        else:
          # Ensure output is created on the correct device with the correct dtype
          output = torch.zeros((1, 1, input.shape[2], input.shape[3]), dtype = input.dtype, device=device)
        return output


def main():
    resnet = Cifar10ResNet()
    module = torch.jit.script(resnet)
    module.save('cifar10_resnet.pt')


if __name__ == '__main__':
    main()
