import torch
import torch.nn as nn

class ColourTransfer(nn.Module):
    """This model is a wrapper around our LinearColourTransfer Model.
    """
    def __init__(self, mixValue = 1.0):
        super(ColourTransfer, self).__init__()
        self.mixValue = mixValue
        self.LinearColourTransfer = LinearColourTransfer()

        # Load weights for LinearColourTransfer
        checkpoint_file = '../checkpoints/colourtransfer_ckpt_600.pth'
        checkpoint = torch.load(checkpoint_file)
        self.LinearColourTransfer.load_state_dict(checkpoint)
        self.LinearColourTransfer.eval()

    def forward(self, input):
        """
        :param input: A torch.Tensor of size 1 x 6 x H x W representing two RGB images
        :return: A torch.Tensor of size 1 x 3 x H x W representing the recoloured image
        """
        # Control self.LinearColourTransfer.mix using mixValue
        self.LinearColourTransfer.mix = self.mixValue

        output = self.LinearColourTransfer(input)

        return output


class LinearColourTransfer(nn.Module):
    """This model transfers the color distribution from one image to another
        using a linear transformation. The variable 'mix' controls how much the
        colours of the first image are changed to look like the second image.
    """
    def __init__(self):
        super(LinearColourTransfer, self).__init__()
        self.linear_layer = nn.Linear(6, 3)
        self.mix = 1.0

    def normalize(self, input):
        """
        :param input: A torch.Tensor of size 1 x 3 x H x W representing the input image
        :return: A torch.Tensor of size 1 x 3 x H x W representing the normalised input image 
        """
        # Compute the mean and std deviation of each color channel 
        means = torch.squeeze(torch.mean(input, (2, 3))) # means has size 1 x 3
        std_dev = torch.squeeze(torch.std(input, (2, 3))) # std_dev has size 1 x 3
        
        # Normalise each channel of input using the mean and std deviation
        output = input.clone()
        output[:, 0, :, :] =  (input[:, 0, :, :] - means[0])*(1/std_dev[0])
        output[:, 1, :, :] =  (input[:, 1, :, :] - means[1])*(1/std_dev[1])
        output[:, 2, :, :] =  (input[:, 2, :, :] - means[2])*(1/std_dev[2])

        return output
        
    def forward(self, input):
        """
        This forward function accepts an input tensor representing two RGB images. The first RGB
        image will be recoloured so that it matches the color distribution of the second.

        :param input: A torch.Tensor of size 1 x 6 x H x W representing two RGB images
        :return: A torch.Tensor of size 1 x 3 x H x W representing the first RGB image,
                    recoloured so that it has the same mean and std deviation as the second image
        """
        # Normalize the first image stored in the tensor
        norm_input = input.clone()
        norm_input[:, 0:3, :, :] = self.normalize(input[:, 0:3, :, :])

        # Reshape the input tensor so it has size [(H * W) x 6]
        # This is the size required by self.linear_layer
        b, c, h, w = norm_input.size()
        reshaped_input = torch.reshape(norm_input, (6, h*w))
        reshaped_input = torch.transpose(reshaped_input, 0, 1)

        # Apply the linear colour transformation to the first image using the
        # colour distribution of the second image
        transformed_input = self.linear_layer(reshaped_input)

        # Reshape the output tensor so it has size [1 x 3 x H x W]
        transformed_input = torch.transpose(transformed_input, 0, 1)
        transformed_input = torch.reshape(transformed_input, (3, h, w))
        transformed_input = torch.unsqueeze(transformed_input, dim = 0)


        # Using .transpose() may have altered the tensor in memory so that it is no longer
        # contiguous, so apply the contiguous() function to fix this
        transformed_input = transformed_input.contiguous()

        # Use the 'mix' variable to control how different the final image will be to the input
        output = (1 - self.mix)*input[:, 0:3, :, :] + (self.mix)*transformed_input

        return output

def main():
    model = ColourTransfer()
    scripted_model = torch.jit.script(model)
    scripted_model.save('colour_transfer.pt')

if __name__ == '__main__':
    main()

